from unittest.mock import MagicMock

from django.contrib.auth import get_user_model
from django.contrib.sessions.backends.file import SessionStore
from django.http import HttpResponse, JsonResponse

from hijack import middleware


class TestHijackRemoteUserMiddleware:
    middleware = middleware.HijackUserMiddleware(lambda x: None)

    def test_process_request__session_not_accessed(self, rf, bob):
        request = rf.get("/")
        request.user = bob
        request.session = SessionStore()
        self.middleware.process_request(request)
        assert not request.session.accessed

    def test_process_request__hijacked(self, rf, bob, alice):
        request = rf.get("/")
        request.user = bob
        request.session = SessionStore()
        request.session["hijack_history"] = [str(alice.pk)]
        self.middleware.process_request(request)
        assert request.user.is_hijacked

    def test_process_request__not_hijacked(self, rf, bob):
        request = rf.get("/")
        request.user = bob
        request.session = SessionStore("some_session_key")
        self.middleware.process_request(request)
        assert not request.user.is_hijacked

    def test_process_request__remote_user__hijacked(self, rf, bob, alice):
        request = rf.get("/")
        request.user = bob
        request.META["REMOTE_USER"] = "eve"
        request.session = SessionStore()
        request.session["hijack_history"] = [str(alice.pk)]
        self.middleware.process_request(request)
        assert request.user.is_hijacked
        assert request.META["REMOTE_USER"] == "bob"

    def test_process_request__remote_user__not_hijacked(self, rf, bob):
        request = rf.get("/")
        request.user = bob
        request.META["REMOTE_USER"] = "eve"
        request.session = SessionStore("some_session_key")
        self.middleware.process_request(request)
        assert not request.user.is_hijacked
        assert request.META["REMOTE_USER"] == "eve"

    def test_process_response__session_not_accessed(self, rf, monkeypatch, bob):
        request = rf.get("/")
        request.user = bob
        request.session = SessionStore()
        response = HttpResponse("")
        assert self.middleware.process_response(request, response) is response
        assert not request.session.accessed

    def test_process_response__html(self, rf, monkeypatch):
        request = rf.get("/")
        request.user = get_user_model()
        request.user.is_hijacked = True
        request.session = SessionStore("some_session_key")
        request.session.accessed = True
        request.META["CSRF_COOKIE"] = "123456"
        render_to_string = MagicMock()
        response = HttpResponse("")
        monkeypatch.setattr("hijack.middleware.render_to_string", render_to_string)
        assert self.middleware.process_response(request, response) is response
        render_to_string.assert_called_once_with(
            "hijack/notification.html",
            {"request": request},
            request=request,
        )

    def test_process_response__non_html(self, rf, monkeypatch):
        request = rf.get("/")
        request.user = get_user_model()
        request.user.is_hijacked = True
        request.session = SessionStore("some_session_key")
        render_to_string = MagicMock()
        response = JsonResponse({})
        monkeypatch.setattr("hijack.middleware.render_to_string", render_to_string)
        assert self.middleware.process_response(request, response) is response
        render_to_string.assert_not_called()

    def test_process_response__content_length(self, rf, monkeypatch):
        request = rf.get("/")
        request.user = get_user_model()
        request.user.is_hijacked = True
        request.session = SessionStore("some_session_key")
        request.session.accessed = True
        request.META["CSRF_COOKIE"] = "123456"
        render_to_string = MagicMock()
        render_to_string.return_value = "HIJACKED"
        response = HttpResponse(b"<body></body>")
        response["Content-Length"] = "13"
        monkeypatch.setattr("hijack.middleware.render_to_string", render_to_string)
        assert self.middleware.process_response(request, response) is response
        render_to_string.assert_called_once_with(
            "hijack/notification.html",
            {"request": request},
            request=request,
        )
        assert response["Content-Length"] == "21"
        assert response.content == b"<body>HIJACKED</body>"

    def test_process_response__insert_before__is_none(self, rf, settings, monkeypatch):
        settings.HIJACK_INSERT_BEFORE = None
        request = rf.get("/")
        request.user = get_user_model()
        request.user.is_hijacked = True
        request.session = SessionStore("some_session_key")
        request.META["CSRF_COOKIE"] = "123456"
        render_to_string = MagicMock()
        monkeypatch.setattr("hijack.middleware.render_to_string", render_to_string)
        response = HttpResponse(b"<body></body>")
        assert self.middleware.process_response(request, response) is response
        render_to_string.assert_not_called()
        assert response.content == b"<body></body>"

    def test_process_response__no_vary_on_user(self, rf, monkeypatch):
        request = rf.get("/")
        request.user = get_user_model()
        request.user.is_hijacked = True
        request.session = SessionStore("some_session_key")
        request.META["CSRF_COOKIE"] = "123456"
        render_to_string = MagicMock()
        response = HttpResponse("")
        monkeypatch.setattr("hijack.middleware.render_to_string", render_to_string)
        assert self.middleware.process_response(request, response) is response
        # the session was not accessed, not need to vary on it
        render_to_string.assert_not_called()



import logging
import warnings

from django.apps import AppConfig
from django.contrib.auth import get_user_model

logger = logging.getLogger(__name__)


class HijackAdminConfig(AppConfig):
    name = "hijack.contrib.admin"
    label = "hijack_admin"

    def ready(self):
        from django.contrib import admin

        from . import HijackUserAdminMixin

        UserModel = get_user_model()
        try:
            UserModelAdmin = type(admin.site._registry[UserModel])
        except KeyError:
            # User model may not be registered, see also:
            # https://github.com/django-hijack/django-hijack/issues/328
            warnings.warn(
                f"{UserModel.__qualname__} is not registered in the default admin."
                f" Please integrate hijack into your user admin manually.",
                UserWarning,
            )
        else:
            if issubclass(UserModelAdmin, HijackUserAdminMixin):
                logger.debug(
                    "UserModelAdmin already is a subclass of HijackUserAdminMixin."
                )
            else:
                admin.site.unregister(UserModel)

                # We create a subclass including the HijackUserAdminMixin but keep the name
                # and module, to keep output form failing checks consistent.
                HijackUserModelAdmin = type(
                    UserModelAdmin.__name__, (HijackUserAdminMixin, UserModelAdmin), {}
                )
                HijackUserModelAdmin.__module__ = UserModelAdmin.__module__

                admin.site.register(UserModel, HijackUserModelAdmin)



import uuid

from django.conf import settings
from django.contrib.auth.models import AbstractUser, UserManager
from django.db import models


class CustomUserManager(UserManager):
    def select_for_update(self):
        raise NotImplementedError


class CustomUser(AbstractUser):
    uuid = models.UUIDField(editable=False, default=uuid.uuid4, primary_key=True)

    objects = CustomUserManager()


class Post(models.Model):
    author = models.ForeignKey(
        settings.AUTH_USER_MODEL, swappable=True, on_delete=models.CASCADE
    )



import pytest
from django.contrib.auth import get_user_model
from django.test import Client


@pytest.fixture()
def alice(db):
    return get_user_model().objects.create(username="alice", is_staff=True)


@pytest.fixture()
def alice_client(alice):
    client = Client()
    client.force_login(alice)
    return client


@pytest.fixture()
def bob(db):
    return get_user_model().objects.create(username="bob", is_staff=False)


@pytest.fixture()
def bob_client(bob):
    client = Client()
    client.force_login(bob)
    return client


@pytest.fixture()
def eve(db):
    return get_user_model().objects.create(username="eve", is_staff=False)


@pytest.fixture()
def eve_client(eve):
    client = Client()
    client.force_login(eve)
    return client


@pytest.fixture()
def other_db_router(settings):
    settings.DATABASE_ROUTERS = ["hijack.tests.test_app.routers.TestAppRouter"]



def superusers_only(*, hijacker, hijacked):
    """Superusers may impersonate any other user."""
    if not hijacked:
        return False
    return hijacked.is_active and hijacker.is_superuser


def superusers_and_staff(*, hijacker, hijacked):
    """
    Superusers and staff members may impersonate other users.

    A superuser may impersonate any other user.
    A staff member may impersonate any user, except another staff member or superuser.
    """
    if not hijacked or not hijacked.is_active:
        return False

    if hijacker.is_superuser:
        return True

    return hijacker.is_staff and not (hijacked.is_staff or hijacked.is_superuser)



from hijack import forms


class TestAsset:
    def test_init(self):
        asset = forms.Asset("path")
        assert asset.path == "path"

    def test_eq(self):
        asset = forms.Asset("path")
        assert asset == "path"
        assert asset == forms.Asset("path")
        assert asset != forms.Asset("other")

    def test_hash(self):
        asset = forms.Asset("path")
        assert hash(asset) == hash("path")

    def test_str(self, settings):
        settings.STORAGES = {
            "default": {"BACKEND": "django.core.files.storage.FileSystemStorage"},
            "staticfiles": {
                "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage"
            },
        }
        asset = forms.Asset("path")
        assert str(asset) == "/static/path"

    def test_absolute_path(self, settings):
        settings.STORAGES = {
            "default": {"BACKEND": "django.core.files.storage.FileSystemStorage"},
            "staticfiles": {
                "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage"
            },
        }
        asset = forms.Asset("path")
        assert asset.absolute_path("path") == "/static/path"
        assert asset.absolute_path("/path") == "/path"
        assert asset.absolute_path("http://path") == "http://path"
        assert asset.absolute_path("https://path") == "https://path"

    def test_repr(self):
        asset = forms.Asset("path")
        assert repr(asset) == "Asset: 'path'"


class TestESM:
    def test_str(self, settings):
        settings.STORAGES = {
            "default": {"BACKEND": "django.core.files.storage.FileSystemStorage"},
            "staticfiles": {
                "BACKEND": "django.contrib.staticfiles.storage.StaticFilesStorage"
            },
        }
        js = forms.ESM("path")
        assert str(js) == '<script src="/static/path" type="module"></script>'



import logging
from unittest.mock import MagicMock

import pytest
from django.contrib.auth.admin import UserAdmin
from django.urls import reverse

from hijack.contrib.admin import HijackUserAdminMixin
from hijack.contrib.admin.apps import HijackAdminConfig
from hijack.tests.test_app.models import CustomUser, Post


class TestHijackUserAdminMixin:
    def test_user_admin(self, admin_client):
        url = reverse("admin:test_app_customuser_changelist")
        response = admin_client.get(url)
        assert response.status_code == 200
        assert b"data-hijack-user" in response.content

    @pytest.fixture()
    def no_user_admin(self):
        from django.contrib import admin

        custom_user_admin = admin.site._registry.pop(CustomUser)
        yield
        admin.site._registry[CustomUser] = custom_user_admin

    @pytest.fixture()
    def custom_hijack_user_admin(self):
        from django.contrib import admin

        CustomUserAdmin = admin.site._registry.pop(CustomUser)
        HijackAdmin = type(UserAdmin.__name__, (HijackUserAdminMixin, UserAdmin), {})
        HijackAdmin.__module__ = UserAdmin.__module__
        admin.site.register(CustomUser, HijackAdmin)
        yield
        admin.site.unregister(CustomUser)
        admin.site._registry[CustomUser] = CustomUserAdmin

    def test_user_admin__unregistered(self, no_user_admin, admin_client):
        with pytest.warns(UserWarning, match="CustomUser is not registered"):
            HijackAdminConfig.ready(None)

    def test_user_admin__custom_subclass(
        self, custom_hijack_user_admin, admin_client, caplog
    ):
        with caplog.at_level(logging.DEBUG):
            HijackAdminConfig.ready(None)
        assert (
            "UserModelAdmin already is a subclass of HijackUserAdminMixin."
            in caplog.text
        )

    def test_related_user(self, admin_client, admin_user):
        url = reverse("admin:test_app_post_changelist")
        Post.objects.create(author=admin_user)
        response = admin_client.get(url)
        assert response.status_code == 200
        assert b"impersonate admin" in response.content

    def test_get_hijack_success_url__obj_absolute_url(self, rf):
        obj = Post()
        obj.get_absolute_url = MagicMock(return_value="/path/to/obj/")
        admin = HijackUserAdminMixin()
        assert admin.get_hijack_success_url(None, obj) == "/path/to/obj/"

    def test_get_hijack_success_url__obj_no_absolute_url(self, rf):
        obj = Post()
        admin = HijackUserAdminMixin()
        assert admin.get_hijack_success_url(None, obj) == "/accounts/profile/"

    def test_get_hijack_success_url__hijack_success_url(self, rf):
        obj = Post()
        obj.get_absolute_url = MagicMock(return_value="/path/to/obj/")
        admin = HijackUserAdminMixin()
        admin.hijack_success_url = "/custom/success/path/"
        assert admin.get_hijack_success_url(None, obj) == "/custom/success/path/"



import contextlib

from django.contrib.auth import user_logged_in
from django.contrib.auth.models import update_last_login
from django.dispatch import Signal

hijack_started = Signal()
hijack_ended = Signal()


@contextlib.contextmanager
def no_update_last_login():
    """Disconnect any signals to ``update_last_login`` and reconnect them on exit."""
    kw = {"receiver": update_last_login}
    kw_id = {"receiver": update_last_login, "dispatch_uid": "update_last_login"}

    was_connected = user_logged_in.disconnect(**kw)
    was_connected_id = not was_connected and user_logged_in.disconnect(**kw_id)
    yield
    # Restore signal if needed
    if was_connected:
        user_logged_in.connect(**kw)
    elif was_connected_id:
        user_logged_in.connect(**kw_id)



from contextlib import contextmanager

from django.contrib.auth import BACKEND_SESSION_KEY, get_user_model, load_backend, login
from django.contrib.auth.mixins import LoginRequiredMixin, UserPassesTestMixin
from django.db import ConnectionRouter, transaction
from django.http import HttpResponseBadRequest, HttpResponseRedirect
from django.shortcuts import get_object_or_404, resolve_url
from django.utils.decorators import method_decorator
from django.utils.http import url_has_allowed_host_and_scheme
from django.utils.module_loading import import_string
from django.views import View
from django.views.decorators.csrf import csrf_protect
from django.views.generic.detail import SingleObjectMixin

from hijack import signals
from hijack.conf import settings


def get_used_backend(request):
    backend_str = request.session[BACKEND_SESSION_KEY]
    backend = load_backend(backend_str)
    return backend


@contextmanager
def keep_session_age(session):
    try:
        session_expiry = session["_session_expiry"]
    except KeyError:
        yield
    else:
        yield
        session["_session_expiry"] = session_expiry


class SuccessUrlMixin:
    redirect_field_name = "next"

    success_url = "/"

    def get_success_url(self):
        url = self.get_redirect_url()
        return url or resolve_url(self.success_url or "/")

    def get_redirect_url(self):
        """Return the user-originating redirect URL if it's safe."""
        redirect_to = self.request.POST.get(
            self.redirect_field_name, self.request.GET.get(self.redirect_field_name, "")
        )
        url_is_safe = url_has_allowed_host_and_scheme(
            url=redirect_to,
            allowed_hosts=self.request.get_host(),
            require_https=self.request.is_secure(),
        )
        return redirect_to if url_is_safe else ""


class LockUserTableMixin:
    def dispatch(self, request, *args, **kwargs):
        write_db = ConnectionRouter().db_for_write(get_user_model())
        # Lock user row to avoid race conditions
        with transaction.atomic(using=write_db):
            get_user_model()._base_manager.select_for_update().get(pk=request.user.pk)
            return super().dispatch(request, *args, **kwargs)


class AcquireUserView(
    LoginRequiredMixin,
    LockUserTableMixin,
    UserPassesTestMixin,
    SuccessUrlMixin,
    SingleObjectMixin,
    View,
):
    model = get_user_model()
    success_url = settings.LOGIN_REDIRECT_URL

    def test_func(self):
        func = import_string(settings.HIJACK_PERMISSION_CHECK)
        return func(hijacker=self.request.user, hijacked=self.get_object())

    def get_object(self, queryset=None):
        return get_object_or_404(self.model, pk=self.request.POST["user_pk"])

    def dispatch(self, request, *args, **kwargs):
        if "user_pk" not in self.request.POST:
            return HttpResponseBadRequest()
        return super().dispatch(request, *args, **kwargs)

    @method_decorator(csrf_protect)
    def post(self, request, *args, **kwargs):
        hijacker = request.user
        hijacked = self.get_object()

        hijack_history = request.session.get("hijack_history", [])
        hijack_history.append(request.user._meta.pk.value_to_string(hijacker))

        backend = get_used_backend(request)
        backend = f"{backend.__module__}.{backend.__class__.__name__}"

        with signals.no_update_last_login(), keep_session_age(request.session):
            login(request, hijacked, backend=backend)

        request.session["hijack_history"] = hijack_history

        signals.hijack_started.send(
            sender=None,
            request=request,
            hijacker=hijacker,
            hijacked=hijacked,
        )
        return HttpResponseRedirect(self.get_success_url())


class ReleaseUserView(
    LoginRequiredMixin,
    LockUserTableMixin,
    UserPassesTestMixin,
    SuccessUrlMixin,
    View,
):
    raise_exception = True

    success_url = settings.LOGOUT_REDIRECT_URL

    def test_func(self):
        return bool(self.request.session.get("hijack_history", []))

    @method_decorator(csrf_protect)
    def post(self, request, *args, **kwargs):
        hijack_history = request.session.get("hijack_history", [])
        hijacked = request.user
        user_pk = hijack_history.pop()
        hijacker = get_object_or_404(get_user_model(), pk=user_pk)
        backend = get_used_backend(request)
        backend = f"{backend.__module__}.{backend.__class__.__name__}"
        with signals.no_update_last_login(), keep_session_age(request.session):
            login(request, hijacker, backend=backend)

        request.session["hijack_history"] = hijack_history

        signals.hijack_ended.send(
            sender=None,
            request=request,
            hijacker=hijacker,
            hijacked=hijacked,
        )
        return HttpResponseRedirect(self.get_success_url())



from django.contrib.auth import get_user_model
from django.http import JsonResponse
from django.views import generic


class UserListView(generic.ListView):
    model = get_user_model()
    template_name = "user_list.html"


class UserDetailView(generic.View):
    def get(self, request, *args, **kwargs):
        return JsonResponse({"username": request.user.username})



import pytest
from django.contrib.auth import get_user_model

from hijack import permissions

superuser = get_user_model()(is_superuser=True)
staff_member = get_user_model()(is_staff=True)
regular_user = get_user_model()()
inactive_user = get_user_model()(is_active=False)


@pytest.mark.parametrize(
    "hijacker, hijacked, has_perm",
    [
        (superuser, None, False),
        (superuser, superuser, True),
        (superuser, staff_member, True),
        (superuser, regular_user, True),
        (superuser, inactive_user, False),
        (staff_member, None, False),
        (staff_member, regular_user, False),
        (staff_member, staff_member, False),
        (staff_member, superuser, False),
        (staff_member, inactive_user, False),
        (regular_user, None, False),
        (regular_user, superuser, False),
        (regular_user, staff_member, False),
        (regular_user, regular_user, False),
        (regular_user, inactive_user, False),
    ],
)
def test_superusers_only(hijacker, hijacked, has_perm):
    assert permissions.superusers_only(hijacker=hijacker, hijacked=hijacked) == has_perm


@pytest.mark.parametrize(
    "hijacker, hijacked, has_perm",
    [
        (superuser, None, False),
        (superuser, superuser, True),
        (superuser, staff_member, True),
        (superuser, regular_user, True),
        (superuser, inactive_user, False),
        (staff_member, None, False),
        (staff_member, regular_user, True),
        (staff_member, staff_member, False),
        (staff_member, superuser, False),
        (staff_member, inactive_user, False),
        (regular_user, None, False),
        (regular_user, superuser, False),
        (regular_user, staff_member, False),
        (regular_user, regular_user, False),
        (regular_user, inactive_user, False),
    ],
)
def test_superusers_and_staff(hijacker, hijacked, has_perm):
    assert (
        permissions.superusers_and_staff(hijacker=hijacker, hijacked=hijacked)
        == has_perm
    )



import copy
import re

from django.template.loader import render_to_string
from django.utils.deprecation import MiddlewareMixin
from django.utils.functional import SimpleLazyObject

from hijack.conf import settings

__all__ = ["HijackUserMiddleware"]

_HTML_TYPES = ("text/html", "application/xhtml+xml")


class HijackUserMiddleware(MiddlewareMixin):
    """Set `is_hijacked` attribute; render and inject notification."""

    @staticmethod
    def setup_user(request, user):
        """Set `is_hijacked` attribute."""
        user.is_hijacked = bool(request.session.get("hijack_history", []))
        return user

    def process_request(self, request):
        """Set `is_hijacked` and override REMOTE_USER header."""
        user = copy.copy(request.user)  # prevent recursion error
        request.user = SimpleLazyObject(lambda: self.setup_user(request, user))
        if "REMOTE_USER" in request.META and request.user.is_hijacked:
            request.META["REMOTE_USER"] = request.user.get_username()

    def process_response(self, request, response):
        """Render hijack notification and inject into HTML response."""
        if request.session.is_empty() or not request.session.accessed:
            # do not touch empty sessions to avoid unnecessary vary on cookie header
            return response

        insert_before = settings.HIJACK_INSERT_BEFORE
        if not getattr(request.user, "is_hijacked", False) or insert_before is None:
            return response

        # Check for responses where the toolbar can't be inserted.
        content_encoding = response.get("Content-Encoding", "")
        content_type = response.get("Content-Type", "").split(";")[0]
        if (
            getattr(response, "streaming", False)
            or "gzip" in content_encoding
            or content_type not in _HTML_TYPES
        ):
            return response

        rendered = render_to_string(
            "hijack/notification.html",
            {"request": request},
            request=request,
        )

        # Insert the toolbar in the response.
        content = response.content.decode(response.charset)
        pattern = re.escape(insert_before)
        bits = re.split(pattern, content, flags=re.IGNORECASE)
        if len(bits) > 1:
            bits[-2] += rendered
            response.content = insert_before.join(bits)
            if "Content-Length" in response:
                response["Content-Length"] = len(response.content)
        return response



from django.contrib import admin
from django.contrib.auth.admin import UserAdmin

from hijack.contrib.admin import HijackUserAdminMixin

from . import models


@admin.register(models.CustomUser)
class CustomUserAdmin(UserAdmin):
    pass


@admin.register(models.Post)
class PostAdmin(HijackUserAdminMixin, admin.ModelAdmin):
    def get_hijack_user(self, obj):
        return obj.author



from django.contrib.auth import user_logged_in
from django.contrib.auth.models import update_last_login

from hijack.signals import no_update_last_login


def test_no_update_last_login(bob, client):
    assert bob.last_login is None
    with no_update_last_login():
        client.force_login(bob)
    assert bob.last_login is None


def test_no_update_last_login__no_dispatch_uid(bob, client):
    assert bob.last_login is None
    with no_update_last_login():
        user_logged_in.connect(update_last_login)
        with no_update_last_login():
            client.force_login(bob)
    assert bob.last_login is None



class TestAppRouter:
    def db_for_read(self, model, **hints):
        return "other"

    def db_for_write(self, model, **hints):
        return "other"

    def allow_relation(self, *args, **kwargs):
        return True

    def allow_syncdb(self, *args, **kwargs):
        return True

    def allow_migrate(self, *args, **kwargs):
        return True



import django
from django import forms
from django.shortcuts import resolve_url
from django.template.loader import render_to_string
from django.utils.translation import gettext_lazy as _

from hijack.conf import settings
from hijack.forms import ESM


class HijackUserAdminMixin:
    """Add hijack button to changelist admin view."""

    hijack_success_url = None
    """Return URL to which one will be forwarded to after hijacking another user."""

    @property
    def media(self):
        return super().media + forms.Media(js=[ESM("hijack/hijack.js")])

    def get_hijack_user(self, obj):
        """
        Return the user based on the current object.

        This method may be overridden to support hijack keys on related objects.
        """
        return obj

    def get_hijack_success_url(self, request, obj):
        """Return URL to which one will be forwarded to after hijacking another user."""
        success_url = settings.LOGIN_REDIRECT_URL
        if self.hijack_success_url:
            success_url = self.hijack_success_url
        elif hasattr(obj, "get_absolute_url"):
            success_url = obj
        return resolve_url(success_url)

    def hijack_button(self, request, obj):
        """
        Render hijack button.

        Should the user only be a related object we include the username in the button
        to ensure deliberate action. However, the name is omitted in the user admin,
        as the table layout suggests that the button targets the current user.
        """
        user = self.get_hijack_user(obj)
        return render_to_string(
            "hijack/contrib/admin/button.html",
            {
                "request": request,
                "another_user": user,
                "username": str(user),
                "is_user_admin": self.model == type(user),
                "next": self.get_hijack_success_url(request, obj),
            },
            request=request,
        )

    def get_changelist_instance(self, request):
        # We inject the request for the CSRF token, see also:
        # https://code.djangoproject.com/ticket/13659
        def hijack_field(obj):
            return self.hijack_button(request, obj)

        hijack_field.short_description = _("impersonate user")

        # we
        list_display = [*self.get_list_display(request), hijack_field]
        # Same as super method, see also:
        # https://github.com/django/django/blob/76c0b32f826469320c59709d31e2f2126dd7c505/django/contrib/admin/options.py#L724-L750
        list_display_links = self.get_list_display_links(request, list_display)
        # Add the action checkboxes if any actions are available.
        if self.get_actions(request):
            list_display = ["action_checkbox", *list_display]
        sortable_by = self.get_sortable_by(request)
        ChangeList = self.get_changelist(request)
        args = [
            request,
            self.model,
            list_display,
            list_display_links,
            self.get_list_filter(request),
            self.date_hierarchy,
            self.get_search_fields(request),
            self.get_list_select_related(request),
            self.list_per_page,
            self.list_max_show_all,
            self.list_editable,
            self,
            sortable_by,
        ]
        if django.VERSION >= (4, 0):
            args.append(self.search_help_text)
        return ChangeList(*args)



from django.conf import settings as django_settings

__all__ = ["settings"]


class LazySettings:
    HIJACK_PERMISSION_CHECK = "hijack.permissions.superusers_only"
    HIJACK_INSERT_BEFORE = "</body>"

    def __getattribute__(self, name):
        try:
            return getattr(django_settings, name)
        except AttributeError:
            return super().__getattribute__(name)


settings = LazySettings()



from django import template
from django.utils.module_loading import import_string

from hijack.conf import settings

register = template.Library()


@register.filter
def can_hijack(hijacker, hijacked):
    """
    Test if a user can hijack another user.

    Usage:

        {% if request.user|can_hijack:another_user %}
          {# The currently authenticated user can hijack the user "another_user". #}
        {% endif %}

    """
    func = import_string(settings.HIJACK_PERMISSION_CHECK)
    return func(hijacker=hijacker, hijacked=hijacked)



from django.urls import reverse

from hijack.templatetags import hijack as templatetags

from .test_app.permissions import allow_all, deny_all


def test_can_hijack__integration(admin_client):
    response = admin_client.get(reverse("user-list"))
    assert response.status_code == 200
    assert b'<button type="submit">impersonate admin</button>' in response.content


def test_can_hijack__import_string(settings):
    settings.HIJACK_PERMISSION_CHECK = (
        f"{allow_all.__module__}.{allow_all.__qualname__}"
    )
    assert templatetags.can_hijack(None, None)

    settings.HIJACK_PERMISSION_CHECK = f"{deny_all.__module__}.{deny_all.__qualname__}"
    assert not templatetags.can_hijack(None, None)



from django.templatetags.static import static
from django.utils.html import format_html, html_safe


@html_safe
class Asset:
    """A generic asset that can be included in a template."""

    def __init__(self, path):
        self.path = path

    def __eq__(self, other):
        return (self.__class__ == other.__class__ and self.path == other.path) or (
            other.__class__ == str and self.path == other
        )

    def __hash__(self):
        return hash(self.path)

    def __str__(self):
        return self.absolute_path(self.path)

    def absolute_path(self, path):
        if path.startswith(("http://", "https://", "/")):
            return path
        return static(path)

    def __repr__(self):
        return f"{type(self).__qualname__}: {self.path!r}"


class ESM(Asset):
    """A JavaScript asset for ECMA Script Modules (ESM)."""

    def __str__(self):
        path = super().__str__()
        template = '<script src="{}" type="module"></script>'
        return format_html(template, self.absolute_path(path))



from django.urls import path

from hijack import views

app_name = "hijack"
urlpatterns = [
    path("acquire/", views.AcquireUserView.as_view(), name="acquire"),
    path("release/", views.ReleaseUserView.as_view(), name="release"),
]



from django.contrib import admin
from django.urls import include, path
from django.views.generic import TemplateView

from . import views

app_name = "test_app"
urlpatterns = [
    path("hijack/", include("hijack.urls", namespace="hijack")),
    path("users/", views.UserListView.as_view(), name="user-list"),
    path("accounts/profile/", views.UserDetailView.as_view(), name="user-detail"),
    path(
        "bye-bye/", TemplateView.as_view(template_name="bye_bye.html"), name="bye-bye"
    ),
    path("admin/", admin.site.urls),
]



"""Settings that need to be set in order to run the tests."""

import os

import django
from django.urls import reverse_lazy

DEBUG = True

SITE_ID = 1

APP_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), "../.."))

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    },
    "other": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    },
}


ROOT_URLCONF = "hijack.tests.test_app.urls"

STATIC_URL = "/static/"
STATIC_ROOT = os.path.join(APP_ROOT, "../app_static")
MEDIA_ROOT = os.path.join(APP_ROOT, "../app_media")
STATICFILES_DIRS = (os.path.join(APP_ROOT, "static"),)

NOSE_ARGS = []

TEMPLATE_DIRS = (os.path.join(APP_ROOT, "tests/test_app/templates"),)

MIDDLEWARE = (
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "hijack.middleware.HijackUserMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
)

TEMPLATE_CONTEXT_PROCESSORS = (
    "django.contrib.auth.context_processors.auth",
    "django.core.context_processors.debug",
    "django.core.context_processors.i18n",
    "django.core.context_processors.media",
    "django.core.context_processors.static",
    "django.core.context_processors.tz",
    "django.contrib.messages.context_processors.messages",
    "django.core.context_processors.request",
)

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": TEMPLATE_DIRS,
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.contrib.auth.context_processors.auth",
                "django.template.context_processors.debug",
                "django.template.context_processors.i18n",
                "django.template.context_processors.media",
                "django.template.context_processors.static",
                "django.template.context_processors.tz",
                "django.contrib.messages.context_processors.messages",
                "django.template.context_processors.request",
            ],
        },
    },
]

EXTERNAL_APPS = [
    "django.contrib.admin",
    "django.contrib.admindocs",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.messages",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "django.contrib.sitemaps",
    "django.contrib.sites",
]

INTERNAL_APPS = [
    "hijack",
    "hijack.contrib.admin",
    "hijack.tests.test_app",
]

INSTALLED_APPS = EXTERNAL_APPS + INTERNAL_APPS

SECRET_KEY = "foobar"  # noqa: S105

LANGUAGE_CODE = "en-us"
USE_I18N = True
USE_TZ = True

LOGOUT_REDIRECT_URL = reverse_lazy("bye-bye")

AUTH_USER_MODEL = "test_app.CustomUser"

if django.VERSION < (4, 0):
    SESSION_SERIALIZER = "django.contrib.sessions.serializers.PickleSerializer"



import datetime
import time
from importlib import import_module
from unittest.mock import MagicMock
from urllib.parse import urlencode, urlunparse

import pytest
from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.sessions.models import Session
from django.db import connections
from django.shortcuts import resolve_url
from django.test import Client
from django.test.utils import CaptureQueriesContext
from django.urls import reverse_lazy
from django.utils import timezone
from django.views import View

from hijack import views


class TestLockUserTableMixin:
    @pytest.mark.django_db(transaction=True)
    def test_dispatch__custom_manager(self, admin_user, rf):
        class LockedView(views.LockUserTableMixin, View):
            pass

        request = rf.get("/")
        request.user = admin_user
        LockedView().dispatch(request)


class TestSuccessUrlMixin:
    def test_get_success_url__path(self):
        view = views.SuccessUrlMixin()
        get_redirect_url = MagicMock()
        get_redirect_url.return_value = ""
        setattr(view, "get_redirect_url", get_redirect_url)
        assert view.get_success_url() == "/"

    def test_get_success_url__pattern(self):
        view = views.SuccessUrlMixin()
        view.success_url = "bye-bye"
        get_redirect_url = MagicMock()
        get_redirect_url.return_value = ""
        setattr(view, "get_redirect_url", get_redirect_url)
        assert view.get_success_url() == "/bye-bye/"


class TestAcquireUserView:
    url = reverse_lazy("hijack:acquire")
    user_detail_url = reverse_lazy("user-detail")

    def test_acquire(self, admin_client, bob):
        assert (
            admin_client.get(self.user_detail_url).content == b'{"username": "admin"}'
        )
        response = admin_client.post(self.url, {"user_pk": bob.pk})
        assert response.status_code == 302
        assert admin_client.get(self.user_detail_url).content == b'{"username": "bob"}'

    def test_acquire__denied(self, eve_client, bob):
        assert eve_client.get(self.user_detail_url).content == b'{"username": "eve"}'
        response = eve_client.post(self.url, {"user_pk": bob.pk})
        assert response.status_code == 403
        assert eve_client.get(self.user_detail_url).content == b'{"username": "eve"}'

    def test_success_url__default(self, admin_client, bob):
        assert (
            admin_client.get(self.user_detail_url).content == b'{"username": "admin"}'
        )
        response = admin_client.post(self.url, {"user_pk": bob.pk})
        assert response.status_code == 302
        assert response["Location"] == "/accounts/profile/"

    def test_success_url__next(self, admin_client, bob):
        assert (
            admin_client.get(self.user_detail_url).content == b'{"username": "admin"}'
        )
        response = admin_client.post(
            self.url,
            {"user_pk": bob.pk, "next": "/somewhere/over/the/rainbow"},
        )
        assert response.status_code == 302
        assert response["Location"] == "/somewhere/over/the/rainbow"

    def test_dispatch__400(self, admin_client):
        response = admin_client.post(self.url)
        assert response.status_code == 400

    def test_unauthenticated(self, client, bob):
        response = client.post(self.url, {"user_pk": bob.pk})
        assert response.status_code == 302
        assert response["Location"] == urlunparse(
            (
                "",
                "",
                resolve_url(settings.LOGIN_URL),
                "",
                urlencode({"next": self.url}, safe="/"),
                "",
            ),
        )

    @pytest.mark.django_db(databases=["default", "other"])
    def test_locking_secondary_db(self, other_db_router):
        # Users are in the "other" db since we're using the other_db_router fixture.
        james = get_user_model().objects.create(username="james")
        frank = get_user_model().objects.create(
            username="frank", is_staff=True, is_superuser=True
        )
        frank_client = Client()
        frank_client.force_login(frank)
        assert (
            frank_client.get(self.user_detail_url).content == b'{"username": "frank"}'
        )

        with CaptureQueriesContext(connections["default"]) as ctx_default:
            with CaptureQueriesContext(connections["other"]) as ctx_other:
                response = frank_client.post(self.url, {"user_pk": james.pk})

        # Ensure that transaction.atomic was routed to the "other" db.
        assert any("SAVEPOINT" in query["sql"] for query in ctx_other.captured_queries)
        # Ensure no queries were made against the "default" db.
        assert len(ctx_default.captured_queries) == 0
        assert response.status_code == 302
        assert (
            frank_client.get(self.user_detail_url).content == b'{"username": "james"}'
        )


class TestReleaseUserView:
    release_url = reverse_lazy("hijack:release")
    acquire_url = reverse_lazy("hijack:acquire")
    user_detail_url = reverse_lazy("user-detail")

    def test_get__405(self, admin_client, alice):
        admin_client.post(self.acquire_url, {"user_pk": alice.pk})
        assert admin_client.get(self.release_url).status_code == 405

    def test_post__not_hijacked(self, admin_client):
        assert admin_client.post(self.release_url).status_code == 403

    @pytest.fixture
    def hijacked_user(self, client, settings):
        engine = import_module(settings.SESSION_ENGINE)
        store = engine.SessionStore()
        store.save()

        return store

    def test_post__hijacked(self, bob_client, alice):
        assert bob_client.get(self.user_detail_url).content == b'{"username": "bob"}'
        session = bob_client.session
        session["hijack_history"] = [str(alice.pk)]
        session.save()

        response = bob_client.post(self.release_url)
        assert response.status_code == 302
        assert not bob_client.session.get("hijack_history")
        assert bob_client.get(self.user_detail_url).content == b'{"username": "alice"}'

    def test_success_url__default(self, bob_client, alice):
        session = bob_client.session
        session["hijack_history"] = [str(alice.pk)]
        session.save()
        assert bob_client.get(self.user_detail_url).content == b'{"username": "bob"}'
        response = bob_client.post(self.release_url)
        assert response.status_code == 302
        assert response["Location"] == "/bye-bye/"

    def test_success_url__next(self, bob_client, alice):
        session = bob_client.session
        session["hijack_history"] = [str(alice.pk)]
        session.save()
        assert bob_client.get(self.user_detail_url).content == b'{"username": "bob"}'
        response = bob_client.post(self.release_url, {"next": "/somewhere/else"})
        assert response.status_code == 302
        assert response["Location"] == "/somewhere/else"

    def test_unauthenticated(self, client, alice):
        response = client.post(self.release_url, {"user_pk": alice.pk})
        assert response.status_code == 403


class TestIntegration:
    acquire_url = reverse_lazy("hijack:acquire")
    release_url = reverse_lazy("hijack:release")

    user_detail_url = reverse_lazy("user-detail")

    def test_acquire_release(self, settings, admin_client, bob, alice):
        settings.HIJACK_PERMISSION_CHECK = "hijack.tests.test_app.permissions.allow_all"
        assert (
            admin_client.get(self.user_detail_url).content == b'{"username": "admin"}'
        )
        response = admin_client.post(self.acquire_url, {"user_pk": bob.pk})
        assert response.status_code == 302
        assert admin_client.get(self.user_detail_url).content == b'{"username": "bob"}'

        response = admin_client.post(self.acquire_url, {"user_pk": alice.pk})
        assert response.status_code == 302
        assert (
            admin_client.get(self.user_detail_url).content == b'{"username": "alice"}'
        )

        response = admin_client.post(self.release_url)
        assert response.status_code == 302
        assert admin_client.get(self.user_detail_url).content == b'{"username": "bob"}'

        response = admin_client.post(self.release_url)
        assert response.status_code == 302
        assert (
            admin_client.get(self.user_detail_url).content == b'{"username": "admin"}'
        )

        response = admin_client.post(self.release_url)
        assert response.status_code == 403
        assert (
            admin_client.get(self.user_detail_url).content == b'{"username": "admin"}'
        )

    def test_keep_session_age(self, admin_client, bob, settings):
        engine = import_module(settings.SESSION_ENGINE)
        session = engine.SessionStore(
            admin_client.cookies[settings.SESSION_COOKIE_NAME].value
        )
        expire_date = timezone.now() + datetime.timedelta(hours=30)
        session.set_expiry(expire_date)
        session.save()
        time.sleep(0.1)
        admin_client.post(self.acquire_url, {"user_pk": bob.pk})
        assert expire_date == Session.objects.get().expire_date
        time.sleep(0.1)
        admin_client.post(self.release_url)
        assert expire_date == Session.objects.get().expire_date
